program iv, eclass
	_on_colon_parse `0'
	local command `"`s(after)'"'
	local 0 `"`s(before)'"'
	syntax varlist [if] [in]

	local ivset `varlist'

	_prefix_command iv `if' `in', : `command'
	
	if "`r(cmdname)'" ~= "runmlwin" {
		display as error "The iv prefix only works with the runmlwin command"
		exit 198
	}
	
	capture unab vars: `r(anything)'
	if `c(rc)' {
		display as error "Invalid varlist"
		exit 198
	}
	tokenize `vars'
	local resp `1'
	macro shift
	local pred `*'
	
	if `:list sizeof ivset' != `:list sizeof pred' {
		display("Variable lists do not match")
		exit 198
	}
	
	local numlevels 1
	gettoken comma tmpstr : command, parse(",") bind
	while strpos("`tmpstr'", "level") != 0 {
		local tmpstr = substr("`tmpstr'", strpos("`tmpstr'", "level") + 5, .)
		local lev = substr("`tmpstr'", 1, strpos("`tmpstr'", "(") - 1)
		* Check this really is a level option
		if `=real("`lev'")' != . {
			if `lev' > `numlevels' {
				local numlevels `lev'
			}
		}
	}
		
	local levargs
	forvalues l = 1/`numlevels' {
		local levargs LEVEL`l'(string) `levargs'
	}
	
	local 0 `command'
	syntax [anything], `levargs' [DISCRETE(string) MCMC(string) CONSTRAINTS(numlist) *]
	forvalues l = 1/`numlevels' {
		gettoken lev`l'id 0 : level`l', parse(":")
		gettoken colon rp`l': 0, parse(":")
	}
	
	if "`rp`numlevels''" == "" {
		display as error "Random variables must be defined at the highest level"
		exit 198
	}
	
	if "`constraints'" ~= "" {
		display as error "Additional constraints not supported"
		exit 198
	}
	
	if "`mcmc'" ~= "" {
		display as error "Estimation procedure not compatible with MCMC"
		exit 198
	}
	
	if "`discrete'" ~= "" {
		display as error "Estimation procedure only works for univariate models"
		exit 198
	}
		
	mata: getbeta("`resp'", "`pred'", "`ivset'");
	
	local count = 1
	local constrnums
	foreach b of varlist `pred' {
		constraint free
		local constrnums `constrnums' `r(free)'
		constraint define `r(free)' [FP1]`b' = `=el(BETA, `count', 1)'
		local ++count
	}
	
	return clear
	
	quietly `command' constraints(`constrnums')
	
	constraint drop `constrnums'
	
	tempname V
	mat `V' = e(V)
	mata: fixse("`resp'", "`pred'", "`ivset'", "`V'")

	ereturn repost V=`V'
	runmlwin, `options'
end

mata:

void getbeta(string scalar yvar, string matrix xvars, string matrix zvars)
{
	X = J(0,0,.);
	Z = J(0,0,.);
	Y = J(0,0,.);
	
	st_view(Y, ., yvar);
	st_view(X, ., xvars);
	st_view(Z, ., zvars);
	
	BETA = lusolve(quadcross(Z, X), quadcross(Z, Y));
	st_matrix("BETA", BETA)
}

void fixse(string scalar yvar, string matrix xvars, string matrix zvars, string scalar Vmat)
{
	LEVIDS = tokens(st_global("e(ivars)") + "  " + st_global("e(level1var)"));
	LEVIDS = LEVIDS[length(LEVIDS)::1];
	numlevs = length(LEVIDS)
	
	Y = J(0,0,.);
	st_view(Y, ., yvar);
	X = J(0,0,.);
	st_view(X, ., xvars);
	Z = J(0,0,.);
	st_view(Z, ., zvars);
	LEVS = J(0,0,.);
	st_view(LEVS, ., invtokens(LEVIDS));
	
	LEVEQNS = J(1, numlevs, "");
	LEVVARS = J(1, numlevs, NULL);
	for (i = 1; i <= numlevs; i++) {
		LEVEQNS[i] = "RP" + strofreal(i);
		LEVVARS[i] = &J(0,0,.)
		st_view(*LEVVARS[i], ., st_local("rp" + strofreal(i)));
	}

	ZXFM1ZT = luinv(quadcross(Z, X)) * Z'
	
	numlevs = length(LEVIDS);
	LEVN = LEVS[., cols(LEVS)];
	
	blocks = uniqrows(LEVN);
	indstart = 1;
	temp = J(cols(X), 0, .)
	for (i = 1; i <= length(blocks); i++) {
		blocksize = sum(LEVN :== blocks[i]);
		indend = indstart + blocksize - 1;
		ZXFM1ZTblock = select(ZXFM1ZT, (LEVN :== blocks[i])');
		Vblock = J(blocksize, blocksize, 0);
		for (j = 1; j <= numlevs; j++) {
			subblocknums = select(LEVS[., j], LEVN :== blocks[i]);
			subblockids = uniqrows(subblocknums);
			rpxblock = select(*(LEVVARS[j]), (LEVN :== blocks[i]));
			subblocks = J(0,0,.);
			for (k = 1; k <= length(subblockids); k++) {
				size = length(select(subblocknums, subblocknums :== subblockids[k]));
				subblocks = blockdiag(subblocks, J(size, size, 1))
			}
			rpmat = st_matrix("e(RP" + strofreal(j) + ")");
			Vblock = Vblock + ((rpxblock * rpmat * rpxblock') :* subblocks);
		}
		
		temp = temp, ZXFM1ZTblock * Vblock
		indstart = indend + 1;
	}
	
	FPV = temp * Z * luinv(quadcross(X, Z));
	V = st_matrix("e(V)");
	V[|1,1 \ cols(X), cols(X)|] = FPV;

	st_replacematrix(Vmat, V);
}

end
